require 'rspec/core/metadata'
require 'method_source'

module RSpec
  module Parser
    module Struct
      TestCase = ::Struct.new(:name, :full_name, :file_path, :file_line, :scoped_id, :code) do
        def self.create(example_instance)
          name = example_instance.description
          full_name = example_instance.description
          file_path = example_instance.metadata[:file_path]
          file_line = example_instance.metadata[:line_number]
          scope_id = example_instance.metadata[:scoped_id]
          file_lines = File.readlines(file_path)
          code = MethodSource.expression_at(file_lines, file_line)
          new(name, full_name, file_path, file_line, scope_id, code)
        end
      end

      Hook = ::Struct.new(:file_path, :file_line, :code) do
        def self.create(hook_instance)
          path, line = *hook_instance.block.source_location
          path = ::RSpec::Core::Metadata.relative_path(path)
          code = hook_instance.block.source
          new(path, line, code)
        end
      end

      Suite = ::Struct.new(:name, :full_name, :file_path, :file_line, :scoped_id, :suites, :examples, :hooks) do
        def to_h
          super.merge(suites: suites.map { |s| s.to_h },
                      examples: examples.map { |e| e.to_h })
        end

        def self.create(example_group_class)
          name = example_group_class.description
          full_name = example_group_class.metadata[:full_description]
          file_path = example_group_class.metadata[:file_path]
          file_line = example_group_class.metadata[:line_number]
          scope_id = example_group_class.metadata[:scoped_id]
          suites = example_group_class.children.map { |g| self.create g }
          examples = example_group_class.examples.map { |e| TestCase.create(e) }
          hooks = example_group_class.hooks
          hooks = [:before, :after].inject({}) do |h_p, p|
            position_hooks = [:context, :example].inject({}) do |h_s, s|
              p_s_hooks = hooks.instance_exec { all_hooks_for(p, s) }
              h_s.merge(s => p_s_hooks.map { |h| Hook.create(h).to_h })
            end
            h_p.merge(p => position_hooks)
          end
          new(name, full_name, file_path, file_line, scope_id, suites, examples, hooks)
        end

        def add_child(other_suite)
          same_name_child = suites.find { |s| s.name == other_suite.name }
          if same_name_child
            # Not allow suites that conflicted with same full paths
            unless same_name_child.examples.empty? or other_suite.examples.empty?
              raise("Conflict: existing same full path suite:#{same_name_child.full_name}")
            end
            other_suite.examples.each { |e| same_name_child.add_example e }
            other_suite.suites.each { |s| same_name_child.add_child s }
          else
            suites << other_suite
          end
        end

        def add_example(example)
          raise('Conflict: existing same full path example!') if examples.any? { |e| e.name == example.name }
          examples << example
        end
      end
    end
  end
end